/*
**  Copyright (c) 2003,2004 Andrea Luzzardi <scox@gentoo.org>
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <pwd.h>
#include <sys/mount.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <dirent.h>

#include "pam.h"
#include "parser.h"
#include "conf.h"
#include "device.h"

extern struct _usb_conf usb_conf;

char *find_proc_directory (void)
{
	static int i=0;
	struct stat proc;
	char *directory = NULL;

	directory = malloc (strlen(usb_conf.proc_basename) + 1);
	memset (directory, '\0', strlen(usb_conf.proc_basename));
	snprintf (directory, strlen(usb_conf.proc_basename), 
		  usb_conf.proc_basename, i);

	if (stat(directory, &proc) == -1) {
		i=0;
		DEBUGP("Directory %s not found\n", directory);
		return NULL;
	}
      	
	i++;
	return directory;
	
}

char *find_proc_file (void)
{
	struct dirent *dir_entry;
	static DIR *dirp = NULL;
	static char *directory = NULL;
	char *file;

	do {
		if (!dirp) {
			if (!(directory = find_proc_directory()))
				return NULL;
			if (!(dirp = opendir(directory)))
				return NULL;
			
			DEBUGP ("Directory %s found\n", directory);
		}
	
		if ((dir_entry = readdir(dirp))) {
			if (strchr(dir_entry->d_name, '.'))
				continue;
			file = insert_before (directory, dir_entry->d_name);
			DEBUGP ("File %s found\n", file);
			return file;
		}

		free (dirp);
		dirp = NULL;

	} while (1);

	return NULL;

}

char *get_private_filename (const char *user, char *mntpoint)
{
	char hostname[32];
	char *file;
	int len;

	if (gethostname (hostname, 32) != 0)
		return NULL;

	len = strlen (mntpoint) + strlen (usb_conf.pwfile) + 
		strlen (user) + strlen (hostname) + 2;

	file = malloc (len + 1);
	memset (file, '\0', len + 1);

	snprintf (file, len, "%s%s%s.%s", mntpoint, usb_conf.pwfile, 
		  user, hostname);

	return file;
}

char *get_public_filename (const char *user)
{
	struct passwd *pw;
	char *home = NULL;
	char *file;
	int len;

	setpwent();
	while ((pw = getpwent())) {
		if (!strncmp ((char *)user, pw->pw_name, strlen(user)) &&
		    !strncmp (pw->pw_name, (char*)user, strlen (pw->pw_name))) {
			home = pw->pw_dir;
			break;
		}
	}
	
	if (!home) {
		DEBUGP ("Cannot retrieve %s user information\n", user);
		return NULL;
	}

	len = strlen (home) + strlen (usb_conf.pwfile) + 
		strlen (usb_conf.pubkey) + 2;

	file = malloc (len);
	memset (file, '\0', len);
	snprintf (file, len-1, "%s%s%s", home, usb_conf.pwfile, usb_conf.pubkey);
	return file;
}

FILE *open_dev_userfile (char *mntpoint, const char *user)
{
	FILE *f;
	char *file;

	if (!(file = get_private_filename (user, mntpoint))) {
		DEBUGP ("Cannot open private file\n");
		return NULL;
	}
	
	DEBUGP ("Opening %s\n", file);

	if (!(f = fopen (file, "r"))) {
		free (file);
		return NULL;
	}

	free (file);
	return f;
}

unsigned long get_mount_opts (void)
{
	char **opts = NULL;
	int i = 0;
	unsigned long mnt_opts = (0xC0ED<<16);

	opts = split (usb_conf.mount_opts, ',');

	while (opts[i]) {
		
		if (!strncmp (opts[i], "ro", strlen(opts[i]))) {
			mnt_opts |= MS_RDONLY;
		} else if (!strncmp (opts[i], "bind", strlen(opts[i]))) {
			mnt_opts |= MS_BIND;
		} else if (!strncmp (opts[i], "sync", strlen(opts[i]))) {
			mnt_opts |= MS_SYNCHRONOUS;
		} else if (!strncmp (opts[i], "remount", strlen(opts[i]))) {
			mnt_opts |= MS_REMOUNT;
		} else if (!strncmp (opts[i], "nosuid", strlen(opts[i]))) {
			mnt_opts |= MS_NOSUID;
		} else if (!strncmp (opts[i], "noexec", strlen(opts[i]))) {
			mnt_opts |= MS_NOEXEC;
		} else if (!strncmp (opts[i], "nodev", strlen(opts[i]))) {
			mnt_opts |= MS_NODEV;
		} else {
			DEBUGP ("Unknown mount option: %s\n", opts[i]);
		}
	
		++i;
	}

	DEBUGP ("Parsed %d mount options (%s)\n", i, usb_conf.mount_opts);

	drop_split (opts);

	return mnt_opts;
}

FILE *device_mount (char *dev_name, const char *user)
{

	FILE *f;
	char **fs = NULL;
	unsigned long mnt_opts = get_mount_opts();
	int i = 0, ret = -1;
	
	fs = split (usb_conf.fs, ',');

	while (fs[i]) {
	
		DEBUGP ("Trying to mount %s on %s using %s\n", dev_name, 
		usb_conf.mntpoint, fs[i]);

		if (!(ret = mount (dev_name, usb_conf.mntpoint, fs[i], mnt_opts, NULL)))
			break;
			
		DEBUGP("mount failed: %s\n", strerror(errno));
		++i;
	}
	
	drop_split (fs);
	
	if (ret == -1) {
		DEBUGP ("Unable to mount %s, tried with %d fs\n", dev_name, i);
		return NULL;
	}
	
	DEBUGP ("Device mounted, trying to open private key\n");

	if (!(f = open_dev_userfile (usb_conf.mntpoint, user))) {
		DEBUGP("Cannot open password file\n");
		umount(usb_conf.mntpoint);
		return NULL;
	}

	DEBUGP("Private key opened\n");
	return f;
}

FILE *try_device_mount (char *dev, const char *user)
{
	
	int try = 0;
	struct stat dev_stat;
	FILE *f;
	char *mntpoint;

	/* strlen ("/dev/sdXN") = 9 */

        /* we got a valid /dev/sdXN name, a devfs entry or a 
	   user forced device*/
      	if (strlen(dev) >= 9 || *(usb_conf.force_device)) 
		return device_mount(dev, user);
	
	if (strlen(dev) != 8) 
		return NULL;
	

	dev = realloc (dev, 10);
	DEBUGP("Trying to guess device name...\n");
	do {
		/* FIXME: Cannot work with devices greater than sda9 */
		dev[8] = try + '1';
		dev[9] = '\0';

		if (usb_conf.check_if_mounted == 1 && 
		    (mntpoint = already_mounted(dev)) &&
		    (f = open_dev_userfile (mntpoint, user))) {
			DEBUGP ("Device already mounted on %s\n", mntpoint);
			return f;
		}

		if ((f = device_mount(dev, user))) {
			return f;
		}
		++try;
	} while (stat(dev, &dev_stat) != -1);
	
	
	return NULL;
}

char *already_mounted (char *dev)
{
	FILE *f;
	char *device = NULL;
	char *mntpoint = NULL;

	if (!(f = fopen ("/proc/mounts", "r"))) {
		DEBUGP ("Cannot open mounts file\n");
		return NULL;
	}

	while ((device = getline_until (f, ' '))) {
		if (!strncmp (device, dev, strlen(device))) {
			mntpoint = getline_until (f, ' ');
			fclose (f);
			free (device);
			return mntpoint;
		}
		free (device);
		free (getline_until (f, '\n'));
	}

	return NULL;
}


FILE *try_device_open (char *dev, const char *user)
{
	char *mntpoint;
	
	if (usb_conf.check_if_mounted == 1 && 
	    (mntpoint = already_mounted(dev))) {
		DEBUGP ("Device already mounted on %s\n", mntpoint);
		return open_dev_userfile (mntpoint, user);
	}
	
	if (strstr (usb_conf.mntpoint, "XXXXXX")) {
		DEBUGP ("Creating temporary mount point...\n");
		if (!(usb_conf.mntpoint = mkdtemp (usb_conf.mntpoint))) {
			DEBUGP ("Cannot get temporary mntpoint (%s)\n", 
				strerror(errno));
			return NULL;
		}
		
		/* As we created the /tmp mount point, we have to delete it */
		usb_conf.delete_mntpoint = 1;
	} else {
		usb_conf.delete_mntpoint = -1;
	}

	DEBUGP ("Using %s as mount point\n", usb_conf.mntpoint);

	return try_device_mount (dev, user);
}

FILE *open_dev (const char *user)
{
	FILE *f;
	FILE *proc;
	int i;
       	char *devnum = NULL, *devname = NULL;
	
	if (*(usb_conf.force_device)) {
		DEBUGP ("Forcing device %s\n", usb_conf.force_device);
		
		if ((f = try_device_open(usb_conf.force_device, user)))
			return f;

		DEBUGP ("Device forcing failed, back to guess mode\n");
	}

       	if (!(proc = fopen ("/proc/partitions", "r")))
		return NULL;

	while ((devnum = getline_until (proc, ' '))) {

	       	for (i=0; i<=1; ++i)
			free (getline_until(proc, ' '));
		devname = getline_until (proc, ' ');

		if (atoi (devnum) != 8) {
			free (devnum);
			free (devname);
			/* if we're not at the end of the line, then get to the next line */
			if (!strchr (devname, '\n')) 
				free (getline_until (proc, '\n'));
			continue;
		}
	

		replace_char (devname, '\n', '\0');
		devname = insert_before ("/dev/", devname);
		DEBUGP("Found a valid device (%s)\n", devname);

		if (!(f = try_device_open(devname, user))) {
			DEBUGP ("invalid device %s\n", devname);
			free (devname);
			free (devnum);
			continue;
		}

		free (devnum);
		free (devname);
		return f;
	}
	DEBUGP("Cannot find any device\n");
	rmdir (usb_conf.mntpoint);
	return NULL;
}

void close_dev (FILE *dev)
{
	fclose (dev);

	if (usb_conf.keep_mounted == 1)
		return;

	umount (usb_conf.mntpoint);
	
	if (usb_conf.delete_mntpoint == -1)
		return;
		
	rmdir (usb_conf.mntpoint);
}







